#pragma once

#include <J3ML/LinearAlgebra.h>

// Default Camera Values
const float YAW = -90.f;
const float PITCH = 0.f;
const float SPEED = 2.5f;
const float SENSITIVITY = 0.1f;
const float ZOOM = 45.0f;

namespace LearnOpenGL
{

    struct CameraInputState
    {
        bool MoveForward;
        bool MoveBackward;
        bool MoveLeft;
        bool MoveRight;
        bool MoveUp;
        bool MoveDown;
    };

    class Camera3D
    {
    public:
        Vector3 Position;
        Vector3 Front;
        Vector3 Up;
        Vector3 Right;
        Vector3 WorldUp;
        float Yaw;
        float Pitch;
        float MovementSpeed;
        float MouseSensitivity;
        float Zoom;

        Camera3D(Vector3 position = Vector3::Zero, Vector3 up = Vector3::Up, float yaw = YAW, float pitch = PITCH)
        : Front{0, 0, -1}, MovementSpeed(SPEED), MouseSensitivity(SENSITIVITY), Zoom(ZOOM)
        {

        }

        Matrix4x4 GetViewMatrix()
        {
            return Matrix4x4::LookAt(Position, Position + Front, Up, Up);
        }

        void UpdateState()
        {
            // calculate the new Front vector
            Vector3 front;
            front.x = std::cos(Math::Radians(Yaw) * cos(Math::Radians(Pitch)));
            front.y = std::sin(Math::Radians(Pitch));
            front.z = std::sin(Math::Radians(Yaw) * std::cos(Math::Radians(Pitch)));
            Front = Vector3::Normalize(front);
            // normalize the vectors, because their length gets closer to 0 the more you look up or down which results in slower movement.
            Right = Vector3::Normalize(Vector3::Cross(Front, WorldUp));
            Up    = Vector3::Normalize(Vector3::Cross(Right, Front));
        }



        void ProcessKeyboard(CameraInputState state, float dt)
        {
            float scalar_velocity = MovementSpeed * dt;
            Vector3 Velocity = Vector3::Zero;

            if (state.MoveForward)
                Velocity += Front * scalar_velocity;
            if (state.MoveBackward)
                Velocity -= Front * scalar_velocity;
            if (state.MoveLeft)
                Velocity -= Right * scalar_velocity;
            if (state.MoveRight)
                Velocity += Right * scalar_velocity;
            if (state.MoveUp)
                Velocity += WorldUp * scalar_velocity;
            if (state.MoveDown)
                Velocity -= WorldUp * scalar_velocity;

            Position += Velocity;

            UpdateState();
        }

        void ProcessMouseMovement(float dx, float dy, bool constrainPitch = true)
        {
            dx *= MouseSensitivity;
            dy *= MouseSensitivity;

            Yaw += dx;
            Pitch += dy;

            if (constrainPitch)
            {
                if (Pitch > 89.0f)
                    Pitch = 89.0f;
                if (Pitch < -89.0f)
                    Pitch = -89.0f;
            }

            UpdateState();
        }

        // processes input received from a mouse scroll-wheel event. Only requires input on the vertical wheel-axis
        void ProcessMouseScroll(float scroll)
        {
            Zoom -= (float)scroll;
            if (Zoom < 1.0f)
                Zoom = 1.0f;
            if (Zoom > 45.0f)
                Zoom = 45.0f;
        }
    public:
    protected:
    private:

    };
}